# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

from dataclasses import dataclass
import os
import platform
from time import monotonic

import plot_layout


@dataclass(frozen=True)
class Metric:
    axis: str
    cumulative: bool


metrics = {
    "peak_nonpaged_pool": Metric("x1y1", False),
    "nonpaged_pool": Metric("x1y1", False),
    "num_page_faults": Metric("x1y2", True),
    "paged_pool": Metric("x1y1", False),
    "peak_paged_pool": Metric("x1y1", False),
    "peak_pagefile": Metric("x1y1", False),
    "peak_wset": Metric("x1y1", False),
    "private": Metric("x1y1", False),
    "rss": Metric("x1y1", False),
    "uss": Metric("x1y1", False),
    "data": Metric("x1y1", False),
    "shared": Metric("x1y1", False),
    "text": Metric("x1y1", False),
    "dirty": Metric("x1y1", False),
    "lib": Metric("x1y1", False),
    "vms": Metric("x1y1", False),
    "other_bytes": Metric("x1y1", True),
    "other_count": Metric("x1y2", True),
    "read_bytes": Metric("x1y1", True),
    "read_chars": Metric("x1y1", True),
    "read_count": Metric("x1y2", True),
    "write_bytes": Metric("x1y1", True),
    "write_chars": Metric("x1y1", True),
    "write_count": Metric("x1y2", True),
    "pfaults": Metric("x1y2", True),
    "pageins": Metric("x1y2", True),
    "minor_faults": Metric("x1y2", True),
    "major_faults": Metric("x1y2", True),
    "pss": Metric("x1y1", False),
    "pss_anon": Metric("x1y1", False),
    "pss_file": Metric("x1y1", False),
    "pss_shmem": Metric("x1y1", False),
    "shared_clean": Metric("x1y1", False),
    "shared_dirty": Metric("x1y1", False),
    "private_clean": Metric("x1y1", False),
    "private_dirty": Metric("x1y1", False),
    "referenced": Metric("x1y1", False),
    "anonymous": Metric("x1y1", False),
    "lazyfree": Metric("x1y1", False),
    "anonhugepages": Metric("x1y1", False),
    "shmempmdmapped": Metric("x1y1", False),
    "filepmdmapped": Metric("x1y1", False),
    "shared_hugetlb": Metric("x1y1", False),
    "private_hugetlb": Metric("x1y1", False),
    "swap": Metric("x1y1", False),
    "swappss": Metric("x1y1", False),
    "locked": Metric("x1y1", False),
}


@dataclass(frozen=True)
class Plot:
    name: str
    title: str
    ylabel: str
    y2label: str
    lines: list[str]


plots = [
    Plot(
        "memory",
        "libtorrent memory usage",
        "Memory Size",
        "",
        [
            "pss",
            "pss_file",
            "pss_anon",
            "rss",
            "dirty",
            "private_dirty",
            "private_clean",
            "lazyfree",
            "anonymous",
            "vms",
            "private",
            "paged_pool",
        ],
    ),
    Plot(
        "vm",
        "libtorrent vm stats",
        "",
        "count",
        [
            "pfaults",
            "pageins",
            "num_page_faults",
            "major_faults",
            "minor_faults",
        ],
    ),
    Plot(
        "read",
        "libtorrent disk I/O (read)",
        "Size",
        "count",
        [
            "read_bytes",
            "read_chars",
            "read_count",
        ],
    ),
    Plot(
        "write",
        "libtorrent disk I/O (write)",
        "Size",
        "count",
        [
            "write_bytes",
            "write_chars",
            "write_count",
        ],
    ),
]

if platform.system() == "Linux":

    def capture_sample(
        pid: int, start_time: float, output: dict[str, list[float]]
    ) -> None:
        try:
            with open(f"/proc/{pid}/smaps_rollup") as f:
                sample = f.read()
            with open(f"/proc/{pid}/stat") as f:
                sample2 = f.read()
            with open(f"/proc/{pid}/io") as f:
                sample3 = f.read()
            timestamp = monotonic() - start_time
        except Exception:
            return

        if "time" not in output:
            time_delta = 0.0
            output["time"] = [timestamp]
        else:
            time_delta = timestamp - output["time"][-1]
            output["time"].append(timestamp)

        for line in sample.split("\n"):
            if "[rollup]" in line:
                continue
            if line.strip() == "":
                continue

            key, value = line.split(":")
            val = int(value.split()[0].strip())
            key = key.strip().lower()

            if key not in output:
                output[key] = [val * 1024]
            else:
                output[key].append(val * 1024)

        stats = sample2.split()

        def add_counter(key: str, val: float) -> None:
            m = metrics[key]
            if key not in output:
                if m.cumulative:
                    output[key + "-raw"] = [val]
                    # we only have a single value, nothing to compare it against
                    val = 0
                output[key] = [val]
            else:
                if m.cumulative:
                    raw_val = val
                    val = (val - output[key + "-raw"][-1]) / time_delta
                    output[key + "-raw"].append(raw_val)
                output[key].append(val)

        add_counter("minor_faults", float(stats[9]))
        add_counter("major_faults", float(stats[11]))

        for line in sample3.split("\n"):
            if line == "":
                continue
            key, val2 = line.split(": ")
            if key == "rchar":
                add_counter("read_chars", float(val2))
            elif key == "wchar":
                add_counter("write_chars", float(val2))
            elif key == "read_bytes":
                add_counter("read_bytes", float(val2))
            elif key == "write_bytes":
                add_counter("write_bytes", float(val2))
            elif key == "syscr":
                add_counter("read_count", float(val2))
            elif key == "syscw":
                add_counter("write_count", float(val2))


# example output:
# 8affffff000-7fffba926000 ---p 00000000 00:00 0  [rollup]
#    Rss:               76932 kB
#    Pss:               17508 kB
#    Pss_Anon:          11376 kB
#    Pss_File:           6101 kB
#    Pss_Shmem:            30 kB
#    Shared_Clean:      65380 kB
#    Shared_Dirty:         88 kB
#    Private_Clean:        80 kB
#    Private_Dirty:     11384 kB
#    Referenced:        76932 kB
#    Anonymous:         11376 kB
#    LazyFree:              0 kB
#    AnonHugePages:         0 kB
#    ShmemPmdMapped:        0 kB
#    FilePmdMapped:         0 kB
#    Shared_Hugetlb:        0 kB
#    Private_Hugetlb:       0 kB
#    Swap:                  0 kB
#    SwapPss:               0 kB
#    Locked:                0 kB

else:
    import psutil

    def capture_sample(
        pid: int, start_time: float, output: dict[str, list[float]]
    ) -> None:
        try:
            p = psutil.Process(pid)
            mem = p.memory_full_info()
            io_cnt = p.io_counters()
            timestamp = monotonic() - start_time
        except Exception:
            return

        if "time" not in output:
            time_delta = 0.0
            output["time"] = [timestamp]
        else:
            time_delta = timestamp - output["time"][-1]
            output["time"].append(timestamp)

        for key in dir(mem):
            if key not in metrics:
                if not key.startswith("_") and key not in [
                    "pagefile",
                    "wset",
                    "count",
                    "index",
                ]:
                    print(f"missing key: {key}")
                continue

            val = getattr(mem, key)

            m = metrics[key]
            if key not in output:
                if m.cumulative:
                    output[key + "-raw"] = [val]
                    # we only have a single value, nothing to compare it against
                    val = 0
                output[key] = [val]
            else:
                if m.cumulative:
                    raw_val = val
                    val = (val - output[key + "-raw"][-1]) / time_delta
                    output[key + "-raw"].append(raw_val)

                output[key].append(val)

        for key in dir(io_cnt):
            if key not in metrics:
                if not key.startswith("_") and key not in [
                    "pagefile",
                    "wset",
                    "count",
                    "index",
                ]:
                    print(f"missing key: {key}")
                continue

            m = metrics[key]
            if key not in output:
                if m.cumulative:
                    output[key + "-raw"] = [val]
                    # we only have a single value, nothing to compare it against
                    val = 0
                output[key] = [val]
            else:
                if m.cumulative:
                    raw_val = val
                    val = (val - output[key + "-raw"][-1]) / time_delta
                    output[key + "-raw"].append(raw_val)

                output[key].append(val)


def print_output_to_file(out: dict[str, list[float]], filename: str) -> list[str]:
    if out == {}:
        return []

    with open(filename, "w+") as stats_output:
        non_zero_keys: set[str] = set()
        non_zero_keys.add("time")
        keys = out.keys()
        for key in keys:
            stats_output.write(f"{key} ")
        stats_output.write("\n")
        idx = 0
        while len(out["time"]) > idx:
            for key in keys:
                stats_output.write(f"{out[key][idx]:.2f} ")
                if out[key][idx] != 0:
                    non_zero_keys.add(key)
            stats_output.write("\n")
            idx += 1
    return [k if k in non_zero_keys else "" for k in keys]


def plot_output(filename: str, keys: list[str]) -> None:
    if "time" not in keys:
        return

    try:
        import matplotlib

        matplotlib.use("Agg")  # headless backend; no display required
        import matplotlib.pyplot as plt
    except ImportError:
        return

    output_dir, in_file = os.path.split(filename)

    # read the stats table written by print_output_to_file: the first line is
    # the column names, the rest are rows of floats.
    with open(filename) as f:
        header = f.readline().split()
        cols: dict[str, list[float]] = {name: [] for name in header}
        for line in f:
            fields = line.split()
            if len(fields) != len(header):
                continue
            for name, value in zip(header, fields):
                cols[name].append(float(value))

    if not cols.get("time"):
        return
    t = cols["time"]

    # print_output_to_file blanks all-zero columns to "" in `keys`; plot only
    # the columns that actually saw a non-zero sample, matching the previous
    # gnuplot output.
    nonzero = {k for k in keys if k}

    mib = 1024 * 1024
    dpi = 100
    # draw the lines a single device pixel wide; matplotlib's ~1.5 pt default
    # is thick for the dense, spiky disk-I/O curves.
    line_width = 72.0 / dpi
    for plot in plots:
        names = [n for n in plot.lines if n in nonzero and n in cols]
        if not names:
            continue

        # count-style (x1y2) metrics go on the left (primary) axis; size
        # (x1y1) metrics go on the right axis, in MB. both axes are always
        # created so the layout is the same across plots, and an axis with no
        # data is left unlabeled. the right (MB) axis tick labels land in the
        # reserved right margin (see subplots_adjust below).
        has_count = any(metrics[n].axis == "x1y2" for n in names)
        has_size = any(metrics[n].axis == "x1y1" for n in names)

        fig, ax = plt.subplots(figsize=(12.0, 7.0))
        ax_mb = ax.twinx()

        # draw colors from a single cycle in plot-line order so the two axes
        # don't both restart at the first color (twinx gives each axis its own
        # cycle otherwise), matching gnuplot's single sequence.
        color_cycle = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        handles = []
        for i, name in enumerate(names):
            m = metrics[name]
            label = name + ("/s" if m.cumulative else "")
            color = color_cycle[i % len(color_cycle)]
            if m.axis == "x1y1":
                # x1y1 metrics are byte counts; show them in MB
                target = ax_mb
                ys = [v / mib for v in cols[name]]
            else:
                target = ax
                ys = cols[name]
            (line,) = target.step(
                t, ys, where="post", label=label, color=color, linewidth=line_width
            )
            handles.append(line)

        ax.set_xlim(left=0)
        ax.set_ylim(bottom=0)
        ax_mb.set_ylim(bottom=0)
        ax.set_xlabel("time (s)")
        ax.set_title(plot.title)
        # label each y-axis only when it has data
        if has_count:
            ax.set_ylabel(plot.y2label)
        if has_size:
            ax_mb.set_ylabel(f"{plot.ylabel} (MB)")
        # vertical grid always; horizontal grid on whichever axis has data
        ax.grid(True, axis="x")
        (ax if has_count else ax_mb).grid(True, axis="y")
        # always top-left; matplotlib's auto-placement does a poor job here
        ax.legend(handles=handles, loc="upper left")

        # pin the plot box to the shared summary-page margins (see plot_layout)
        # so these line up with the piece-pass-order and latency plots.
        fig.subplots_adjust(
            left=plot_layout.BOX_LEFT,
            right=plot_layout.BOX_RIGHT,
            bottom=0.09,
            top=0.93,
        )
        fig.savefig(f"{output_dir}/{in_file}-{plot.name}.png", dpi=dpi)
        plt.close(fig)



#!/usr/bin/env python3


import os
import sys

import libtorrent

if len(sys.argv) < 3:
    print("usage make_torrent.py file tracker-url")
    sys.exit(1)

input = os.path.abspath(sys.argv[1])

fs = []

parent_input = os.path.split(input)[0]

# if we have a single file, use it because os.walk does not work on a single files
if os.path.isfile(input):
    size = os.path.getsize(input)
    fs.append(libtorrent.create_file_entry(input, size))

for root, dirs, files in os.walk(input):
    # skip directories starting with .
    if os.path.split(root)[1][0] == ".":
        continue

    for f in files:
        # skip files starting with .
        if f[0] == ".":
            continue

        # skip thumbs.db on windows
        if f == "Thumbs.db":
            continue

        fname = os.path.join(root[len(parent_input) + 1 :], f)
        size = os.path.getsize(os.path.join(parent_input, fname))
        print("%10d kiB  %s" % (size / 1024, fname))
        fs.append(libtorrent.create_file_entry(fname, size))

if len(fs) == 0:
    print("no files added")
    sys.exit(1)

t = libtorrent.create_torrent(fs, 0, 4 * 1024 * 1024)

t.add_tracker(sys.argv[2])
t.set_creator("libtorrent %s" % libtorrent.__version__)


def log_piece(written_piece: int) -> None:
    sys.stdout.write(".")


libtorrent.set_piece_hashes(t, parent_input, log_piece)
sys.stdout.write("\n")

with open("out.torrent", "wb+") as f_handle:
    f_handle.write(libtorrent.bencode(t.generate()))



import ipaddress
from typing import List
from typing import Tuple
import unittest

import libtorrent as lt


def export(ipf: lt.ip_filter) -> Tuple[List[Tuple[str, str]], List[Tuple[str, str]]]:
    def norm(addr: str) -> str:
        return str(ipaddress.ip_address(addr))

    def norm_list(filter_list: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        return [(norm(lo), norm(hi)) for lo, hi in filter_list]

    v4, v6 = ipf.export_filter()
    return (norm_list(v4), norm_list(v6))


class IpFilterTest(unittest.TestCase):
    maxDiff = None

    def test_empty(self) -> None:
        ipf = lt.ip_filter()
        self.assertEqual(ipf.access("0.1.2.3"), 0)
        self.assertEqual(ipf.access("::123"), 0)

        self.assertEqual(
            export(ipf),
            (
                [("0.0.0.0", "255.255.255.255")],
                [("::", "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff")],
            ),
        )

    def test_with_rule_ip4(self) -> None:
        ipf = lt.ip_filter()
        ipf.add_rule("0.0.0.0", "0.255.255.255", 123)
        self.assertEqual(ipf.access("0.1.2.3"), 123)
        self.assertEqual(ipf.access("1.2.3.4"), 0)
        self.assertEqual(ipf.access("::123"), 0)

        self.assertEqual(
            export(ipf),
            (
                [("0.0.0.0", "0.255.255.255"), ("1.0.0.0", "255.255.255.255")],
                [("::", "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff")],
            ),
        )

    def test_with_rule_ip6(self) -> None:
        ipf = lt.ip_filter()
        ipf.add_rule("::", "::ffff", 123)
        self.assertEqual(ipf.access("::123"), 123)
        self.assertEqual(ipf.access("1.2.3.4"), 0)
        self.assertEqual(ipf.access("::1:0"), 0)

        self.assertEqual(
            export(ipf),
            (
                [("0.0.0.0", "255.255.255.255")],
                [
                    ("::", "::ffff"),
                    ("::1:0", "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff"),
                ],
            ),
        )

    def test_export(self) -> None:
        ipf = lt.ip_filter()
        self.assertEqual(
            export(ipf),
            (
                [("0.0.0.0", "255.255.255.255")],
                [("::", "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff")],
            ),
        )

        ipf = lt.ip_filter()
        ipf.add_rule("0.0.0.0", "0.255.255.255", 123)
        ipf.add_rule("::", "::ffff", 456)
        self.assertEqual(
            export(ipf),
            (
                [("0.0.0.0", "0.255.255.255"), ("1.0.0.0", "255.255.255.255")],
                [
                    ("::", "::ffff"),
                    ("::1:0", "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff"),
                ],
            ),
        )



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

from __future__ import print_function

import glob
import os
import sys

# usage: parse_peer_log <path-to-libtorrent-peer-logs>

log_files = []

for p in glob.iglob(os.path.join(sys.argv[1], '*.log')):
    name = os.path.split(p)[1]
    if name == 'main_session.log':
        continue
    print(name)
    f = open(p, 'r')
    out_file = p + '.dat'
    log_files.append(out_file)
    out = open(out_file, 'w+')

    uploaded_blocks = 0
    downloaded_blocks = 0

    for line in f:
        t = line.split(': ')[0].split('.')[0]
        log_line = False
        if ' ==> PIECE' in line:
            uploaded_blocks += 1
            log_line = True

        if ' <== PIECE' in line:
            downloaded_blocks += 1
            log_line = True

        if log_line:
            print('%s\t%d\t%d' % (t, uploaded_blocks, downloaded_blocks), file=out)

    out.close()
    f.close()

out = open('peers.gnuplot', 'wb')
print("set term png size 1200,700", file=out)
print('set xrange [0:*]', file=out)
print('set xlabel "time"', file=out)
print('set ylabel "blocks"', file=out)
print('set key box', file=out)
print('set xdata time', file=out)
print('set timefmt "%H:%M:%S"', file=out)
print('set title "uploaded blocks"', file=out)
print('set output "peers_upload.png"', file=out)
print('plot', end=' ', file=out)
first = True
for n in log_files:
    if not first:
        print(',', end=' ', file=out)
    first = False
    print(' "%s" using 1:2 title "%s" with steps' % (n, os.path.split(n)[1].split('.log')[0]), end=' ', file=out)
print('', file=out)

print('set title "downloaded blocks"', file=out)
print('set output "peers_download.png"', file=out)
print('plot', end=' ', file=out)
first = True
for n in log_files:
    if not first:
        print(',', end=' ', file=out)
    first = False
    print(' "%s" using 1:3 title "%s" with steps' % (n, os.path.split(n)[1].split('.log')[0]), end=' ', file=out)
print('', file=out)
out.close()

os.system('gnuplot peers.gnuplot')



import tempfile
from typing import Any
from typing import Dict
from typing import TYPE_CHECKING
import unittest

import libtorrent as lt

from . import lib

if TYPE_CHECKING:
    from libtorrent import load_torrent_limits
else:
    load_torrent_limits = dict


def filter(atp: Dict[str, Any]) -> Dict[str, Any]:
    if lt.api_version >= 3:
        del atp["info_hash"]
    if lt.api_version >= 2:
        del atp["url"]
    return atp


class ParseMagnetTest(unittest.TestCase):
    def setUp(self) -> None:
        self.info_hash_sha1 = lib.get_random_bytes(20).hex()
        self.info_hash_sha256 = lib.get_random_bytes(32).hex()

    def test_parse_to_atp(self) -> None:
        uri = f"magnet:?xt=urn:btih:{self.info_hash_sha1}"
        atp = lt.parse_magnet_uri(uri)
        if lt.api_version < 3:
            self.assertEqual(str(atp.info_hash).lower(), self.info_hash_sha1)
        self.assertEqual(str(atp.info_hashes.v1).lower(), self.info_hash_sha1)

    def test_parse_to_atp_error(self) -> None:
        with self.assertRaises(RuntimeError):
            lt.parse_magnet_uri("magnet:?")

    def test_parse_dict_deprecated(self) -> None:
        uri = f"magnet:?xt=urn:btih:{self.info_hash_sha1}"
        with self.assertWarns(DeprecationWarning):
            lt.parse_magnet_uri_dict(uri)

    def test_parse_dict_sha1(self) -> None:
        uri = (
            f"magnet:?xt=urn:btih:{self.info_hash_sha1}&"
            "dn=test.txt&"
            "tr=http://example.com/tr&"
            "ws=http://example.com/ws&"
            "so=0-2,4&"
            "x.pe=0.1.2.3:4567&"
            "dht=1.2.3.4:5678"
        )
        with self.assertWarns(DeprecationWarning):
            params = lt.parse_magnet_uri_dict(uri)
        self.assertEqual(
            params,
            filter(
                {
                    "dht_nodes": [("1.2.3.4", 5678)],
                    "flags": lt.torrent_flags.default_flags
                    | lt.torrent_flags.default_dont_download,
                    "info_hash": bytes.fromhex(self.info_hash_sha1),
                    "info_hashes": bytes.fromhex(self.info_hash_sha1),
                    "name": "test.txt",
                    "save_path": "",
                    "storage_mode": lt.storage_mode_t.storage_mode_sparse,
                    "trackers": ["http://example.com/tr"],
                    "url": "",
                }
            ),
        )

        # The dict is intended to be usable as argument to session.add_torrent()
        session = lt.session(lib.get_isolated_settings())
        with tempfile.TemporaryDirectory() as path:
            params["save_path"] = path
            with self.assertWarns(DeprecationWarning):
                handle = session.add_torrent(params)
            self.assertEqual(str(handle.info_hashes().v1), self.info_hash_sha1)
            self.assertEqual(handle.status().name, "test.txt")
            self.assertEqual(
                [t["url"] for t in handle.trackers()], ["http://example.com/tr"]
            )
            # self.assertEqual(handle.url_seeds(), ["http://example.com/ws"])
            # self.assertEqual(handle.file_priorities(), [4, 4, 4, 0, 4])
            # Can't test peers or dht

    @unittest.skip("need to parse more params")
    def test_parse_dict_sha1_broken(self) -> None:
        uri = (
            f"magnet:?xt=urn:btih:{self.info_hash_sha1}&"
            "dn=test.txt&"
            "tr=http://example.com/tr&"
            "ws=http://example.com/ws&"
            "so=0-2,4&"
            "x.pe=0.1.2.3:4567&"
            "dht=1.2.3.4:5678"
        )
        params = lt.parse_magnet_uri_dict(uri)
        self.assertEqual(
            params,
            filter(
                {
                    "dht_nodes": [("1.2.3.4", 5678)],
                    "file_priorities": [4, 4, 4, 0, 4],
                    "flags": lt.torrent_flags.default_flags
                    | lt.torrent_flags.default_dont_download,
                    "info_hash": bytes.fromhex(self.info_hash_sha1),
                    "info_hashes": bytes.fromhex(self.info_hash_sha1),
                    "name": "test.txt",
                    "save_path": "",
                    "storage_mode": lt.storage_mode_t.storage_mode_sparse,
                    "trackers": ["http://example.com/tr"],
                    "url": "",
                    "url_seeds": ["http://example.com/ws"],
                }
            ),
        )

        # The dict is intended to be usable as argument to session.add_torrent()
        session = lt.session(lib.get_isolated_settings())
        with tempfile.TemporaryDirectory() as path:
            params["save_path"] = path
            handle = session.add_torrent(params)
            self.assertEqual(str(handle.info_hashes().v1), self.info_hash_sha1)
            self.assertEqual(handle.name(), "test.txt")
            self.assertEqual(
                [t["url"] for t in handle.trackers()], ["http://example.com/tr"]
            )
            self.assertEqual(handle.url_seeds(), ["http://example.com/ws"])
            self.assertEqual(handle.file_priorities(), [4, 4, 4, 0, 4])
            # Can't test peers or dht

    def test_parse_dict_sha256(self) -> None:
        uri = (
            f"magnet:?xt=urn:btmh:1220{self.info_hash_sha256}&"
            "dn=test.txt&"
            "tr=http://example.com/tr&"
            "ws=http://example.com/ws&"
            "so=0-2,4&"
            "x.pe=0.1.2.3:4567&"
            "dht=1.2.3.4:5678"
        )
        with self.assertWarns(DeprecationWarning):
            params = lt.parse_magnet_uri_dict(uri)
        self.assertEqual(
            params,
            filter(
                {
                    "dht_nodes": [("1.2.3.4", 5678)],
                    "flags": lt.torrent_flags.default_flags
                    | lt.torrent_flags.default_dont_download,
                    "info_hash": bytes.fromhex(self.info_hash_sha256)[:20],
                    "info_hashes": bytes.fromhex(self.info_hash_sha256),
                    "name": "test.txt",
                    "save_path": "",
                    "storage_mode": lt.storage_mode_t.storage_mode_sparse,
                    "trackers": ["http://example.com/tr"],
                    "url": "",
                }
            ),
        )

        # The dict is intended to be usable as argument to session.add_torrent()
        session = lt.session(lib.get_isolated_settings())
        with tempfile.TemporaryDirectory() as path:
            params["save_path"] = path
            with self.assertWarns(DeprecationWarning):
                handle = session.add_torrent(params)
            self.assertEqual(str(handle.info_hashes().v2), self.info_hash_sha256)
            self.assertEqual(handle.status().name, "test.txt")
            self.assertEqual(
                [t["url"] for t in handle.trackers()], ["http://example.com/tr"]
            )
            # self.assertEqual(handle.url_seeds(), ["http://example.com/ws"])
            # self.assertEqual(handle.file_priorities(), [4, 4, 4, 0, 4])
            # Can't test peers or dht

    @unittest.skip("need to parse more params")
    def test_parse_dict_sha256_broken(self) -> None:
        uri = (
            f"magnet:?xt=urn:btmh:1220{self.info_hash_sha256}&"
            "dn=test.txt&"
            "tr=http://example.com/tr&"
            "ws=http://example.com/ws&"
            "so=0-2,4&"
            "x.pe=0.1.2.3:4567&"
            "dht=1.2.3.4:5678"
        )
        params = lt.parse_magnet_uri_dict(uri)
        self.assertEqual(
            params,
            filter(
                {
                    "dht_nodes": [("1.2.3.4", 5678)],
                    "file_priorities": [4, 4, 4, 0, 4],
                    "flags": lt.torrent_flags.default_flags
                    | lt.torrent_flags.default_dont_download,
                    "info_hash": bytes.fromhex(self.info_hash_sha256)[:20],
                    "info_hashes": bytes.fromhex(self.info_hash_sha256),
                    "name": "test.txt",
                    "peers": [("0.1.2.3", 4567)],
                    "save_path": "",
                    "storage_mode": lt.storage_mode_t.storage_mode_sparse,
                    "trackers": ["http://example.com/tr"],
                    "url": "",
                    "url_seeds": "http://example.com/ws",
                }
            ),
        )

        # The dict is intended to be usable as argument to session.add_torrent()
        session = lt.session(lib.get_isolated_settings())
        with tempfile.TemporaryDirectory() as path:
            params["save_path"] = path
            handle = session.add_torrent(params)
            self.assertEqual(
                str(handle.info_hashes().v2),
                self.info_hash_sha256,
            )
            self.assertEqual(handle.name(), "test.txt")
            self.assertEqual(
                [t["url"] for t in handle.trackers()], ["http://example.com/tr"]
            )
            self.assertEqual(handle.url_seeds(), ["http://example.com/ws"])
            self.assertEqual(handle.file_priorities(), [4, 4, 4, 0, 4])
            # Can't test peers or dht

    def test_parse_dict_error(self) -> None:
        with self.assertWarns(DeprecationWarning):
            with self.assertRaises(RuntimeError):
                lt.parse_magnet_uri_dict("magnet:?")


class AddMagnetUriTest(unittest.TestCase):
    def setUp(self) -> None:
        self.session = lt.session(lib.get_isolated_settings())
        self.dir = tempfile.TemporaryDirectory()
        self.info_hash_sha1 = lib.get_random_bytes(20).hex()

    def tearDown(self) -> None:
        lib.cleanup_with_windows_fix(self.dir, timeout=5)

    def test_error(self) -> None:
        if lt.api_version < 2:
            with self.assertWarns(DeprecationWarning):
                with self.assertRaises(RuntimeError):
                    lt.add_magnet_uri(
                        self.session,
                        "magnet:?",
                        load_torrent_limits({"save_path": "."}),
                    )

    def test_add(self) -> None:
        if lt.api_version < 2:
            uri = f"magnet:?xt=urn:btih:{self.info_hash_sha1}"
            with self.assertWarns(DeprecationWarning):
                handle = lt.add_magnet_uri(
                    self.session, uri, load_torrent_limits({"save_path": self.dir.name})
                )
            self.assertEqual(str(handle.info_hashes().v1), self.info_hash_sha1)


class MakeMagnetUriTest(unittest.TestCase):
    def setUp(self) -> None:
        self.info_hash_sha1 = lib.get_random_bytes(20).hex()

    def test_torrent_info(self) -> None:
        ti = lt.torrent_info(lt.sha1_hash(bytes.fromhex(self.info_hash_sha1)))
        uri = lt.make_magnet_uri(ti)
        self.assertEqual(uri, f"magnet:?xt=urn:btih:{self.info_hash_sha1}")

    def test_torrent_handle(self) -> None:
        atp = lt.add_torrent_params()
        atp.info_hashes = lt.info_hash_t(
            lt.sha1_hash(bytes.fromhex(self.info_hash_sha1))
        )
        session = lt.session(lib.get_isolated_settings())
        with tempfile.TemporaryDirectory() as path:
            atp.save_path = path
            handle = session.add_torrent(atp)
            uri = lt.make_magnet_uri(handle)
        self.assertEqual(uri, f"magnet:?xt=urn:btih:{self.info_hash_sha1}")



#!/usr/bin/env python3
import glob
import subprocess
import os
import re

reportdir = "reports/"

class bcolors:
	HEADER = '\033[95m'
	OKBLUE = '\033[94m'
	OKGREEN = '\033[92m'
	WARNING = '\033[93m'
	FAIL = '\033[91m'
	ENDC = '\033[0m'
	BOLD = '\033[1m'
	UNDERLINE = '\033[4m'


print("Testing crashfiles")

FNULL = open(os.devnull, "w")
crashfiles = []
crashfiles.extend(glob.glob("*"))
pattern = re.compile("^(leak-|timeout-|crash-)\w+$")

filecounter = 1

FNULL = open(os.devnull, 'w')

for filename in crashfiles:

	if not pattern.match(filename):
		continue

	fuzzer_retval = subprocess.call(["./check-input.sh", filename, "batchmode"], stdout=FNULL, stderr=subprocess.STDOUT)
	
	if fuzzer_retval == 0:
		print(bcolors.FAIL, "[", filecounter, "]", filename,"- not reproducable", bcolors.ENDC)
	else:
		print(bcolors.OKGREEN, "[", filecounter, "]", filename, "- reproducable", bcolors.ENDC)

	filecounter = filecounter + 1



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import os

file_header = '''/*

Copyright (c) 2017-2022, Arvid Norberg
Copyright (c) 2017-2018, Steven Siloti
Copyright (c) 2020, Alden Torres
All rights reserved.

You may use, distribute and modify this code under the terms of the BSD license,
see LICENSE file.
*/

#ifndef TORRENT_FWD_HPP
#define TORRENT_FWD_HPP

#include "libtorrent/config.hpp"

namespace libtorrent {
'''

file_footer = '''

	using file_layout = file_storage;

}

namespace lt = libtorrent;

#endif // TORRENT_FWD_HPP
'''

classes = os.popen(
    r'git grep -E "(TORRENT_EXPORT|TORRENT_DEPRECATED_EXPORT|^TORRENT_[A-Z0-9]+_NAMESPACE)"').read().split('\n')


def print_classes(out, classes, keyword):
    current_file = ''

    # [(file, decl), ...]
    classes = [(x.split(':')[0].strip(), ':'.join(x.split(':')[1:]).strip()) for x in classes]

    # we only care about header files
    # ignore the forward header itself, that's the one we're generating
    # also ignore any header in the aux_ directory, those are private
    classes = [x for x in classes if x[0].endswith('.hpp') and not x[0].endswith('/fwd.hpp') and '/aux_/' not in x[0]]

    namespaces = ['TORRENT_VERSION_NAMESPACE_4',
                  'TORRENT_VERSION_NAMESPACE_4_END',
                  'TORRENT_VERSION_NAMESPACE_3',
                  'TORRENT_VERSION_NAMESPACE_3_END',
                  'TORRENT_VERSION_NAMESPACE_2',
                  'TORRENT_VERSION_NAMESPACE_2_END',
                  'TORRENT_CRYPTO_NAMESPACE',
                  'TORRENT_CRYPTO_NAMESPACE_END']

    # only include classes with the right kind of export
    classes = [
        x for x in classes if x[1] in namespaces or (
            x[1].split(' ')[0] in [
                'class',
                'struct'] and x[1].split(' ')[1] == keyword)]

    # collapse empty namespaces
    classes2 = []
    skip = 0
    for i in range(len(classes)):
        if skip > 0:
            skip -= 1
            continue
        if classes[i][1] in namespaces \
                and len(classes) > i + 1 \
                and classes[i + 1][1] == ('%s_END' % classes[i][1]):
            skip = 1
        else:
            classes2.append(classes[i])

    classes = classes2

    idx = -1
    for line in classes:
        idx += 1
        this_file = line[0]
        decl = line[1].split(' ')

        content = ''
        if this_file != current_file:
            out.write('\n// ' + this_file + '\n')
        current_file = this_file
        if len(decl) > 2 and decl[0] in ['struct', 'class']:
            decl = decl[0] + ' ' + decl[2]
            if not decl.endswith(';'):
                decl += ';'
            content = decl + '\n'
        else:
            content = line[1] + '\n'

        if 'kademlia' in this_file:
            out.write('namespace dht {\n')
            out.write(content)
            out.write('}\n')
        else:
            out.write(content)


try:
    os.remove('include/libtorrent/fwd.hpp')
except FileNotFoundError:
    pass

with open('include/libtorrent/fwd.hpp', 'w+') as f:
    f.write(file_header)

    print_classes(f, classes, 'TORRENT_EXPORT')

    f.write('\n#if TORRENT_ABI_VERSION <= 2\n')

    print_classes(f, classes, 'TORRENT_DEPRECATED_EXPORT')

    f.write('\n#endif // TORRENT_ABI_VERSION')

    f.write(file_footer)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import sys
import os

gnuplot_scripts = []


def gen_stats_gnuplot(name, y, lines):

    global gnuplot_scripts

    stat = open(sys.argv[1])
    line = stat.readline()
    while 'minute:' not in line:
        line = stat.readline()

    names = line.strip().split(':')
    counter = 1
    for i in names:
        print('%d: %s' % (counter, i))
        counter += 1

    out = open('%s.gnuplot' % name, 'w+')
    out.write('''
set term png size 1200,700 small
set output "%s.png"
set title "%s"
set ylabel "%s"
set xlabel "time (minutes)"
plot ''' % (name, name.strip('_'), y))
    first = True
    for i in lines:
        if not first:
            out.write(', \\\n')
        first = False
        out.write('"%s" using 1:%d title "%s" with lines' % (sys.argv[1], names.index(i) + 1, i))
    out.write('\n')

    out.write('''set terminal postscript
set output "%s.ps"
replot
''' % (name))
    out.close()
    gnuplot_scripts += [name]


gen_stats_gnuplot('dht_routing_table_size', 'nodes', ['active nodes', 'passive nodes', 'confirmed nodes'])
gen_stats_gnuplot('dht_tracker_table_size', '', ['num torrents', 'num peers'])
gen_stats_gnuplot('dht_announces', 'messages per minute', ['announces per min', 'failed announces per min'])
gen_stats_gnuplot('dht_clients',
                  'messages per minute',
                  ['total msgs per min',
                   'az msgs per min',
                   'ut msgs per min',
                   'lt msgs per min',
                   'mp msgs per min',
                   'gr msgs per min'])
gen_stats_gnuplot('dht_rate', 'bytes per second', ['bytes in per sec', 'bytes out per sec'])
gen_stats_gnuplot('dht_errors', 'messages per minute', ['error replies sent', 'error queries recvd'])

for i in gnuplot_scripts:
    os.system('gnuplot %s.gnuplot' % i)



#!/usr/bin/env python
# -*- coding: utf-8 -*- #

AUTHOR = 'Paul-Louis Ageneau'
COPYRIGHT_YEAR = 2021
SITENAME = 'libdatachannel'
SITEURL = ''

PATH = 'content'
STATIC_PATHS = ['images']

THEME = 'theme'

TIMEZONE = 'Europe/Paris'
DEFAULT_LANG = 'en'

GITHUB_URL = "https://github.com/paullouisageneau/libdatachannel"

# Feed generation is usually not desired when developing
FEED_ALL_ATOM = None
CATEGORY_FEED_ATOM = None
TRANSLATION_FEED_ATOM = None
AUTHOR_FEED_ATOM = None
AUTHOR_FEED_RSS = None

# Blogroll
#
#LINKS = (('Pelican', 'https://getpelican.com/'),
#         ('Python.org', 'https://www.python.org/'),
#         ('Jinja2', 'https://palletsprojects.com/p/jinja/'),
#         ('You can modify those links in your config file', '#'),)

# Social widget
#SOCIAL = (('You can add links in your config file', '#'),
#          ('Another social link', '#'),)

DEFAULT_PAGINATION = False

# Uncomment following line if you want document-relative URLs when developing
#RELATIVE_URLS = True

MARKDOWN = {
    'extension_configs': {
        'markdown.extensions.codehilite': {'css_class': 'highlight'},
        'markdown.extensions.extra': {},
        'markdown.extensions.meta': {},
        'markdown.extensions.fenced_code': {},
    },
    'tab_length': 2,
}


import unittest

import libtorrent as lt


class GenerateFingerprintTest(unittest.TestCase):
    def test_generate(self) -> None:
        # full version
        self.assertEqual(
            lt.generate_fingerprint_bytes(b"AB", 1, 2, 3, 4),
            b"-AB1234-",
        )

        # long name
        with self.assertRaises(ValueError):
            lt.generate_fingerprint_bytes(b"ABCD", 1, 2, 3, 4),

        # short name
        with self.assertRaises(ValueError):
            lt.generate_fingerprint_bytes(b"A", 1, 2, 3, 4)

        # major only
        self.assertEqual(
            lt.generate_fingerprint_bytes(b"AB", 1),
            b"-AB1000-",
        )

        # high versions
        self.assertEqual(
            lt.generate_fingerprint_bytes(b"AB", 1000, 2000, 3000, 4000),
            b"-AB\x1f\x07\xef\xd7-",
        )

        # version < 0
        with self.assertRaises(ValueError):
            lt.generate_fingerprint_bytes(b"AB", -1, -1, -1, -1)

    @unittest.skip("https://github.com/arvidn/libtorrent/issues/5988")
    def test_deprecations(self) -> None:
        with self.assertWarns(DeprecationWarning):
            lt.generate_fingerprint("ABCD", 1, 2, 3, 4)

    def test_generate_str(self) -> None:
        # full version
        self.assertEqual(lt.generate_fingerprint("AB", 1, 2, 3, 4), "-AB1234-")

        # short name
        with self.assertRaises(ValueError):
            self.assertEqual(lt.generate_fingerprint("A", 1, 2, 3, 4), "---1234-")

        # version < 0
        with self.assertRaises(ValueError):
            self.assertEqual(lt.generate_fingerprint("AB", 1, 2, -1, -1), "-AB1200-")


class FingerprintTest(unittest.TestCase):
    def test_fingerprint(self) -> None:
        if lt.api_version < 2:
            with self.assertWarns(DeprecationWarning):
                fprint = lt.fingerprint("AB", 1, 2, 3, 4)
            self.assertEqual(str(fprint), "-AB1234-")
            self.assertEqual(fprint.major_version, 1)
            self.assertEqual(fprint.minor_version, 2)
            self.assertEqual(fprint.revision_version, 3)
            self.assertEqual(fprint.tag_version, 4)

            # short names behave differently
            with self.assertWarns(DeprecationWarning):
                with self.assertRaises(ValueError):
                    fprint = lt.fingerprint("A", 1, 2, 3, 4)



#!/usr/bin/env python3

import glob
import os
import sys

paths = [
    'test/*.cpp',
    'src/*.cpp',
    'src/kademlia/*.cpp',
    'include/libtorrent/*.hpp',
    'include/libtorrent/kademlia/*.hpp',
    'include/libtorrent/aux_/*.hpp',
    'include/libtorrent/extensions/*.hpp']

os.system('(cd .. ; ctags %s 2>/dev/null)' % ' '.join(paths))

files = []

for p in paths:
    files.extend(glob.glob(os.path.join('..', p)))

items = []

# keeps 20 non-comment lines, used to get more context around
# todo-items
context = []

priority_count = [0, 0, 0, 0, 0]


def html_sanitize(s):
    ret = ''
    for i in s:
        if i == '<':
            ret += '&lt;'
        elif i == '>':
            ret += '&gt;'
        elif i == '&':
            ret += '&amp;'
        else:
            ret += i
    return ret


for f in files:
    h = open(f)

    state = ''
    line_no = 0
    context_lines = 0

    for orig_line in h:
        line_no += 1
        line = orig_line.strip()
        if 'TODO:' in line and line.startswith('//'):
            line = line.split('TODO:')[1].strip()
            state = 'todo'
            items.append({})
            items[-1]['location'] = '%s:%d' % (f, line_no)
            items[-1]['priority'] = 0
            if line[0] in '0123456789':
                items[-1]['priority'] = int(line[0])
                if int(line[0]) > 5:
                    print('priority too high: ' + line)
                    sys.exit(1)

                line = line[1:].strip()
            items[-1]['todo'] = line
            prio = items[-1]['priority']
            if prio >= 0 and prio <= 4:
                priority_count[prio] += 1
            continue

        if state == '':
            context.append(html_sanitize(orig_line))
            if len(context) > 20:
                context.pop(0)
            continue

        if state == 'todo':
            if line.strip().startswith('//'):
                items[-1]['todo'] += '\n'
                items[-1]['todo'] += line[2:].strip()
            else:
                state = 'context'
                items[-1]['context'] = ''.join(context) + \
                    '<div style="background: #ffff00" width="100%">' + html_sanitize(orig_line) + '</div>'
                context_lines = 1

                context.append(html_sanitize(orig_line))
                if len(context) > 20:
                    context.pop(0)
            continue

        if state == 'context':
            items[-1]['context'] += html_sanitize(orig_line)
            context_lines += 1

            context.append(html_sanitize(orig_line))
            if len(context) > 20:
                context.pop(0)
            if context_lines > 30:
                state = ''

    h.close()

items.sort(key=lambda x: x['priority'], reverse=True)

# for i in items:
#   print('\n\n', i['todo'], '\n')
#   print(i['location'], '\n')
#   print('prio: ', i['priority'], '\n')
#   if 'context' in i:
#       print(i['context'], '\n')

out = open('todo.html', 'w+')
out.write('''<html><head>
<script type="text/javascript">
/* <![CDATA[ */
	var expanded = -1
	function expand(id) {
		if (expanded != -1) {
			var ctx = document.getElementById(expanded);
			ctx.style.display = "none";
			// if we're expanding the field that's already
			// expanded, just collapse it
			var no_expand = id == expanded;
			expanded = -1;
			if (no_expand) return;
		}
		var ctx = document.getElementById(id);
		ctx.style.display = "table-row";
		expanded = id;
	}
/* ]]> */
</script>

</head><body>
<h1>libtorrent todo-list</h1>
<span style="color: #f00">%d urgent</span>
<span style="color: #f77">%d important</span>
<span style="color: #3c3">%d relevant</span>
<span style="color: #77f">%d feasible</span>
<span style="color: #999">%d notes</span>
          <table width="100%%" border="1" style="border-collapse: collapse;">''' # noqa
          % (priority_count[4], priority_count[3], priority_count[2], priority_count[1], priority_count[0]))

prio_colors = ['#ccc', '#ccf', '#cfc', '#fcc', '#f44']

index = 0
for i in items:
    if 'context' not in i:
        i['context'] = ''
    out.write(('<tr style="background: %s"><td>relevance&nbsp;%d</td>'
               '<td><a href="javascript:expand(%d)">%s</a></td><td>%s</td></tr>')
              % (prio_colors[i['priority']], i['priority'], index, i['location'], i['todo'].replace('\n', ' ')))

    out.write(
        ('<tr id="%d" style="display: none;" colspan="3"><td colspan="3"><h2>%s</h2><h4>%s</h4>'
         '<pre style="background: #f6f6f6; border: solid 1px #ddd;">%s</pre></td></tr>') %
        (index, i['todo'], i['location'], i['context']))
    index += 1

out.write('</table></body></html>')
out.close()



import os
import tempfile
from typing import AnyStr
from typing import cast
from typing import List
from typing import TYPE_CHECKING
import unittest

import libtorrent as lt

from . import lib
from . import tdummy

if TYPE_CHECKING:
    from libtorrent import _Entry
    from libtorrent import load_torrent_limits
    from libtorrent import TorrentFileDict
    from libtorrent import web_seed_entry
else:
    _Entry, TorrentFileDict, load_torrent_limits = dict, dict, dict


class FileEntryTest(unittest.TestCase):
    def test_attributes(self) -> None:
        if lt.api_version < 2:
            with self.assertWarns(DeprecationWarning):
                fe = lt.file_entry()

            fe.path = os.path.join("path", "file.txt")
            path = fe.path
            self.assertEqual(path, os.path.join("path", "file.txt"))

            fe.symlink_path = os.path.join("path", "other.txt")
            self.assertEqual(fe.symlink_path, os.path.join("path", "other.txt"))

            fe.filehash = lt.sha1_hash(b"a" * 20)
            self.assertEqual(fe.filehash, lt.sha1_hash(b"a" * 20))

            fe.mtime = 123456
            self.assertEqual(fe.mtime, 123456)

            with self.assertWarns(DeprecationWarning):
                self.assertFalse(fe.pad_file)
            with self.assertWarns(DeprecationWarning):
                self.assertFalse(fe.executable_attribute)
            with self.assertWarns(DeprecationWarning):
                self.assertFalse(fe.hidden_attribute)
            with self.assertWarns(DeprecationWarning):
                self.assertFalse(fe.symlink_attribute)

            with self.assertWarns(DeprecationWarning):
                self.assertEqual(fe.offset, 0)
            with self.assertWarns(DeprecationWarning):
                self.assertEqual(fe.size, 0)


class InfoHashTest(unittest.TestCase):
    def test_sha1(self) -> None:
        sha1 = lt.sha1_hash(lib.get_random_bytes(20))

        ih = lt.info_hash_t(sha1)
        self.assertEqual(ih.get(lt.protocol_version.V1), sha1)
        self.assertEqual(ih.get(lt.protocol_version.V2), lt.sha1_hash())
        self.assertEqual(ih.get_best(), sha1)
        self.assertTrue(ih.has(lt.protocol_version.V1))
        self.assertFalse(ih.has(lt.protocol_version.V2))
        self.assertTrue(ih.has_v1())
        self.assertFalse(ih.has_v2())
        self.assertEqual(ih.v1, sha1)
        self.assertEqual(ih.v2, lt.sha256_hash())

        self.assertEqual(ih, lt.info_hash_t(sha1))
        self.assertEqual(hash(ih), hash(lt.info_hash_t(sha1)))
        self.assertNotEqual(ih, lt.info_hash_t())

    def test_sha256(self) -> None:
        sha256 = lt.sha256_hash(lib.get_random_bytes(32))
        sha256_trunc = lt.sha1_hash(sha256.to_bytes()[:20])

        ih = lt.info_hash_t(sha256)
        self.assertEqual(ih.get(lt.protocol_version.V1), lt.sha1_hash())
        self.assertEqual(ih.get(lt.protocol_version.V2), sha256_trunc)
        self.assertEqual(ih.get_best(), sha256_trunc)
        self.assertFalse(ih.has(lt.protocol_version.V1))
        self.assertTrue(ih.has(lt.protocol_version.V2))
        self.assertFalse(ih.has_v1())
        self.assertTrue(ih.has_v2())
        self.assertEqual(ih.v1, lt.sha1_hash())
        self.assertEqual(ih.v2, sha256)

        self.assertEqual(ih, lt.info_hash_t(sha256))
        self.assertEqual(hash(ih), hash(lt.info_hash_t(sha256)))
        self.assertNotEqual(ih, lt.info_hash_t())

    def test_dual(self) -> None:
        sha1 = lt.sha1_hash(lib.get_random_bytes(20))
        sha256 = lt.sha256_hash(lib.get_random_bytes(32))
        sha256_trunc = lt.sha1_hash(sha256.to_bytes()[:20])

        ih = lt.info_hash_t(sha1, sha256)
        self.assertEqual(ih.get(lt.protocol_version.V1), sha1)
        self.assertEqual(ih.get(lt.protocol_version.V2), sha256_trunc)
        self.assertEqual(ih.get_best(), sha256_trunc)
        self.assertTrue(ih.has(lt.protocol_version.V1))
        self.assertTrue(ih.has(lt.protocol_version.V2))
        self.assertTrue(ih.has_v1())
        self.assertTrue(ih.has_v2())
        self.assertEqual(ih.v1, sha1)
        self.assertEqual(ih.v2, sha256)

        self.assertEqual(ih, lt.info_hash_t(sha1, sha256))
        self.assertEqual(hash(ih), hash(lt.info_hash_t(sha1, sha256)))
        self.assertNotEqual(ih, lt.info_hash_t())


class ConstructorTest(unittest.TestCase):
    def do_test_filename(self, name: AnyStr) -> None:
        entry = cast(
            "_Entry",
            {
                b"info": {
                    b"name": b"test.txt",
                    b"piece length": 16384,
                    b"pieces": lib.get_random_bytes(20),
                    b"length": 1024,
                }
            },
        )
        data = lt.bencode(entry)
        with tempfile.TemporaryDirectory() as tempdir_str:
            tempdir: AnyStr
            if isinstance(name, str):
                tempdir = tempdir_str
            else:
                tempdir = os.fsencode(tempdir_str)
            path = os.path.join(tempdir, name)
            with open(path, mode="wb") as fp:
                fp.write(data)
            ti = lt.torrent_info(path)
        self.assertEqual(ti.name(), "test.txt")

    def test_filename(self) -> None:
        # ascii str
        self.do_test_filename("test.torrent")

        # non-ascii str
        self.do_test_filename("test-\u1234.torrent")

    @unittest.skip("https://github.com/arvidn/libtorrent/issues/5984")
    @lib.uses_non_unicode_paths()
    def test_filename_non_unicode_str(self) -> None:
        # non-unicode str
        self.do_test_filename(os.fsdecode(b"test-\xff.torrent"))

    def test_dict(self) -> None:
        entry = TorrentFileDict(
            {
                b"info": {
                    b"name": b"test.txt",
                    b"piece length": 16384,
                    b"pieces": lib.get_random_bytes(20),
                    b"length": 1024,
                }
            }
        )
        ti = lt.torrent_info(entry)
        self.assertEqual(ti.name(), "test.txt")

    def test_dict_with_str_keys(self) -> None:
        with self.assertWarns(DeprecationWarning):
            ti = lt.torrent_info(
                TorrentFileDict(
                    {
                        "info": {
                            "name": "test.txt",
                            "length": 1024,
                            "piece length": 16384,
                            "pieces": "aaaaaaaaaaaaaaaaaaaa",
                        }
                    }
                )
            )
        self.assertEqual(ti.name(), "test.txt")

    def test_copy_constructor(self) -> None:
        ti = tdummy.get_default().torrent_info()
        other = lt.torrent_info(ti)
        self.assertEqual(ti.info_hashes(), other.info_hashes())

    def test_info_hash(self) -> None:
        sha1 = lt.sha1_hash(lib.get_random_bytes(20))
        ti = lt.torrent_info(lt.info_hash_t(sha1))
        self.assertEqual(ti.info_hashes(), lt.info_hash_t(sha1))

    def test_sha1_hash(self) -> None:
        sha1 = lt.sha1_hash(lib.get_random_bytes(20))
        ti = lt.torrent_info(sha1)
        self.assertEqual(ti.info_hash(), sha1)


class ConstructorWithLimitsTest(unittest.TestCase):
    def test_load_decode_depth_limit(self) -> None:
        with self.assertRaises(RuntimeError):
            lt.torrent_info(
                TorrentFileDict(
                    {
                        b"test": {b"test": {b"test": {b"test": {b"test": {}}}}},
                        b"info": {
                            b"name": b"test_torrent",
                            b"length": 1234,
                            b"piece length": 16 * 1024,
                            b"pieces": b"aaaaaaaaaaaaaaaaaaaa",
                        },
                    }
                ),
                load_torrent_limits({"max_decode_depth": 1}),
            )

    def test_load_max_pieces_limit(self) -> None:
        with self.assertRaises(RuntimeError):
            lt.torrent_info(
                TorrentFileDict(
                    {
                        b"info": {
                            b"name": b"test_torrent",
                            b"length": 1234000,
                            b"piece length": 16 * 1024,
                            b"pieces": b"aaaaaaaaaaaaaaaaaaaa",
                        }
                    }
                ),
                load_torrent_limits({"max_pieces": 1}),
            )

    def test_load_max_buffer_size_limit(self) -> None:
        with self.assertRaises(RuntimeError):
            lt.torrent_info(
                TorrentFileDict(
                    {
                        b"info": {
                            b"name": b"test_torrent",
                            b"length": 1234000,
                            b"piece length": 16 * 1024,
                            b"pieces": b"aaaaaaaaaaaaaaaaaaaa",
                        }
                    }
                ),
                load_torrent_limits({"max_buffer_size": 1}),
            )


class FieldTest(unittest.TestCase):
    def setUp(self) -> None:
        self.ti = tdummy.get_default().torrent_info()

    def test_trackers(self) -> None:
        self.ti.add_tracker("http://example0.com", 0, lt.tracker_source.source_client)
        with self.assertRaises(TypeError):
            self.ti.add_tracker(
                b"http://example1.com",  # type: ignore
                1,
                int(lt.tracker_source.source_torrent),
            )

        self.assertEqual([t.url for t in self.ti.trackers()], ["http://example0.com"])
        self.assertEqual([t.tier for t in self.ti.trackers()], [0])
        self.assertEqual(
            [t.source for t in self.ti.trackers()], [lt.tracker_source.source_client]
        )

    def test_add_tracker_defaults(self) -> None:
        self.ti.add_tracker("http://example.com")

        self.assertEqual([t.url for t in self.ti.trackers()], ["http://example.com"])
        self.assertEqual([t.tier for t in self.ti.trackers()], [0])
        self.assertEqual(
            [t.source for t in self.ti.trackers()], [lt.tracker_source.source_client]
        )

    def test_web_seeds(self) -> None:
        # empty, and ascii str
        web_seeds: list[web_seed_entry] = [
            {"url": "http://example1.com", "auth": ""},
            {"url": "http://example2.com", "auth": "password"},
        ]
        self.ti.set_web_seeds(web_seeds)
        self.assertEqual(self.ti.web_seeds(), web_seeds)

        def do_test(auth: AnyStr, auth_check: str) -> None:
            new_web_seeds: list[web_seed_entry] = [
                {"url": "http://example.com", "auth": auth}
            ]
            self.ti.set_web_seeds(new_web_seeds)
            self.assertEqual(
                self.ti.web_seeds(), [{"url": "http://example.com", "auth": auth_check}]
            )

        # ascii bytes
        do_test(b"password", "password")

        # non-ascii str
        do_test("\u1234", "\u1234")

        # non-ascii bytes
        do_test("\u1234".encode(), "\u1234")

        # non-unicode
        with self.assertRaises(ValueError):
            do_test(b"\xff", "dummy")

    @unittest.skip("need to implement default empty auth")
    def test_web_seeds_default_auth(self) -> None:
        web_seeds: list[web_seed_entry] = [{"url": "http://example.com"}]
        self.ti.set_web_seeds(web_seeds)
        self.assertEqual(self.ti.web_seeds(), web_seeds)

    def test_add_url_seed(self) -> None:
        self.ti.add_url_seed("http://example1.com")
        self.ti.add_url_seed("http://example2.com", "password", [("header", "value")])

    def test_add_http_seed(self) -> None:
        with self.assertWarns(DeprecationWarning):
            self.ti.add_http_seed("http://example1.com")
        with self.assertWarns(DeprecationWarning):
            self.ti.add_http_seed(
                "http://example2.com", "password", [("header", "value")]
            )

    def test_name(self) -> None:
        def generate(name: bytes) -> lt.torrent_info:
            return lt.torrent_info(
                TorrentFileDict(
                    {
                        b"info": {
                            b"name": name,
                            b"pieces": lib.get_random_bytes(20),
                            b"piece length": 16384,
                            b"length": 1024,
                        }
                    }
                )
            )

        # ascii
        self.assertEqual(generate(b"test.txt").name(), "test.txt")

        # non-ascii
        self.assertEqual(generate("\u1234.txt".encode()).name(), "\u1234.txt")

        # non-unicode
        self.assertEqual(generate(b"\xff.txt").name(), "_.txt")

    def test_creator(self) -> None:
        def generate(creator: bytes) -> lt.torrent_info:
            return lt.torrent_info(
                TorrentFileDict(
                    {
                        b"created by": creator,
                        b"info": {
                            b"name": b"test.txt",
                            b"pieces": lib.get_random_bytes(20),
                            b"piece length": 16384,
                            b"length": 1024,
                        },
                    }
                )
            )

        # ascii
        self.assertEqual(generate(b"test").creator(), "test")

        # non-ascii
        self.assertEqual(generate("test-\u1234".encode()).creator(), "test-\u1234")

        # non-unicode
        self.assertEqual(generate(b"test-\xff").creator(), "test-_")

    def test_comment(self) -> None:
        def generate(comment: bytes) -> lt.torrent_info:
            return lt.torrent_info(
                TorrentFileDict(
                    {
                        b"comment": comment,
                        b"info": {
                            b"name": b"test.txt",
                            b"pieces": lib.get_random_bytes(20),
                            b"piece length": 16384,
                            b"length": 1024,
                        },
                    }
                )
            )

        # ascii
        self.assertEqual(generate(b"test").comment(), "test")

        # non-ascii
        self.assertEqual(generate("\u1234".encode()).comment(), "\u1234")

        # non-unicode
        self.assertEqual(generate(b"test-\xff").comment(), "test-_")

    def test_collections(self) -> None:
        def generate(collection: bytes) -> lt.torrent_info:
            return lt.torrent_info(
                TorrentFileDict(
                    {
                        b"collections": [collection],
                        b"info": {
                            b"name": b"test.txt",
                            b"pieces": lib.get_random_bytes(20),
                            b"piece length": 16384,
                            b"length": 1024,
                        },
                    }
                )
            )

        # ascii
        self.assertEqual(generate(b"test").collections(), ["test"])

        # non-ascii
        self.assertEqual(generate("\u1234".encode()).collections(), ["\u1234"])

        # non-unicode
        with self.assertRaises(ValueError):
            generate(b"\xff").collections()

    def test_hash_for_piece_invalid(self) -> None:
        ti = lt.torrent_info(
            TorrentFileDict(
                {
                    b"info": {
                        b"name": b"test.txt",
                        b"pieces": lib.get_random_bytes(20),
                        b"piece length": 16384,
                        b"length": 1024,
                    }
                }
            )
        )
        with self.assertRaises(IndexError):
            ti.hash_for_piece(1)
        with self.assertRaises(IndexError):
            ti.hash_for_piece(-1)

    def test_info_hash(self) -> None:
        self.assertFalse(self.ti.info_hash().is_all_zeros())
        self.assertEqual(self.ti.info_hashes(), lt.info_hash_t(self.ti.info_hash()))

    def test_is_i2p(self) -> None:
        ti = lt.torrent_info(
            TorrentFileDict(
                {
                    b"announce": b"http://example.i2p",
                    b"info": {
                        b"name": b"test.txt",
                        b"pieces": lib.get_random_bytes(20),
                        b"piece length": 16384,
                        b"length": 1024,
                    },
                }
            )
        )
        self.assertTrue(ti.is_i2p())

    def test_normal_fixed_fields(self) -> None:
        piece = lib.get_random_bytes(20)
        similar = lib.get_random_bytes(20)
        info = cast(
            "_Entry",
            {
                b"name": b"test.txt",
                b"pieces": piece,
                b"piece length": 16384,
                b"length": 1024,
                b"similar": [similar],
            },
        )
        ti = lt.torrent_info(TorrentFileDict({b"creation date": 12345, b"info": info}))

        self.assertEqual(ti.total_size(), 1024)
        self.assertEqual(ti.piece_length(), 16384)
        self.assertEqual(ti.num_pieces(), 1)
        with self.assertRaises(IndexError):
            ti.piece_size(-1)
        self.assertEqual(ti.piece_size(0), 1024)
        with self.assertRaises(IndexError):
            ti.piece_size(1)
        self.assertEqual(ti.hash_for_piece(0), piece)
        self.assertEqual(ti.num_files(), 1)
        self.assertTrue(ti.is_valid())
        self.assertFalse(ti.priv())
        self.assertEqual(ti.info_section(), lt.bencode(info))
        self.assertEqual(ti.creation_date(), 12345)
        self.assertEqual(ti.similar_torrents(), [lt.sha1_hash(similar)])

    def test_metadata(self) -> None:
        with self.assertWarns(DeprecationWarning):
            metadata = self.ti.metadata()
        self.assertEqual(metadata, self.ti.info_section())
        with self.assertWarns(DeprecationWarning):
            size = self.ti.metadata_size()
        self.assertEqual(size, len(metadata))

    def test_merkle_tree(self) -> None:
        with self.assertWarns(DeprecationWarning):
            self.assertFalse(self.ti.is_merkle_torrent())
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(self.ti.merkle_tree(), [])

        tree: List[bytes] = []
        with self.assertWarns(DeprecationWarning):
            self.ti.set_merkle_tree(tree)
        with self.assertWarns(DeprecationWarning):
            self.assertFalse(self.ti.is_merkle_torrent())
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(self.ti.merkle_tree(), tree)

    def test_nodes(self) -> None:
        self.ti.add_node("node1", 1234)
        self.assertEqual(self.ti.nodes(), [("node1", 1234)])

    def test_rename_file(self) -> None:
        # ascii str
        self.ti.rename_file(0, "ascii-str.txt")
        self.assertEqual(self.ti.files().file_path(0), "ascii-str.txt")

        # ascii bytes
        self.ti.rename_file(0, b"ascii-bytes.txt")  # type: ignore
        self.assertEqual(self.ti.files().file_path(0), "ascii-bytes.txt")

        # non-ascii str
        self.ti.rename_file(0, "\u1234-str.txt")
        self.assertEqual(self.ti.files().file_path(0), "\u1234-str.txt")

        # non-ascii bytes
        self.ti.rename_file(0, "\u1234-bytes.txt".encode())  # type: ignore
        self.assertEqual(self.ti.files().file_path(0), "\u1234-bytes.txt")

        # non-unicode
        self.ti.rename_file(0, b"\xff.txt")  # type: ignore
        with self.assertRaises(ValueError):
            self.ti.files().file_path(0)

    def test_files_and_remap(self) -> None:
        ti = lt.torrent_info(
            TorrentFileDict(
                {
                    b"info": {
                        b"name": b"test.txt",
                        b"pieces": lib.get_random_bytes(20),
                        b"piece length": 16384,
                        b"length": 1024,
                    },
                }
            )
        )

        self.assertEqual(ti.name(), "test.txt")
        self.assertEqual(ti.layout().file_path(0), "test.txt")
        self.assertEqual(ti.orig_files().file_path(0), "test.txt")

        fs = lt.file_storage()
        fs.add_file(os.path.join("path", "remapped1.txt"), 512)
        fs.add_file(os.path.join("path", "remapped2.txt"), 512)

        ti.remap_files(fs)

        self.assertEqual(ti.name(), "path")
        self.assertEqual(ti.files().num_files(), 2)
        self.assertEqual(ti.files().file_path(0), os.path.join("path", "remapped1.txt"))
        self.assertEqual(ti.files().file_path(1), os.path.join("path", "remapped2.txt"))

        self.assertEqual(ti.orig_files().num_files(), 1)
        self.assertEqual(ti.orig_files().file_path(0), "test.txt")

    def test_map(self) -> None:
        pr = self.ti.map_file(0, 100, 200)
        self.assertEqual(pr.piece, 0)
        self.assertEqual(pr.start, 100)
        self.assertEqual(pr.length, 200)

        slices = self.ti.map_block(0, 100, 200)
        self.assertEqual(len(slices), 1)
        file_slice = slices[0]
        self.assertEqual(file_slice.file_index, 0)
        self.assertEqual(file_slice.offset, 100)
        self.assertEqual(file_slice.size, 200)

    def test_file_at(self) -> None:
        if lt.api_version < 2:
            with self.assertWarns(DeprecationWarning):
                fe = self.ti.file_at(0)
            self.assertEqual(fe.path, self.ti.layout().file_path(0))
